Here are examples of queries and the corresponding answers for similar functions:
{{ examples }}

Note that the query could be interpreted as a combination of several independent requests.
{{ parallel_queries }}
Based on these examples, generate {{ number }} diverse query and answer pairs for the function `{{ func_name }}`.
The detailed function description is the following:
{{ func_desc }}
{{ format_inst }}
Now please generate {{ number }} diverse query and answer pairs following the above format.