# Copyright 2023-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, List, TypeVar, Union

GenerateOutput = List[Union[str, None]]

if TYPE_CHECKING:
    from numpy import floating
    from numpy.typing import NDArray

    GenericFloat = TypeVar("GenericFloat", bound=floating[Any])

    HiddenState = NDArray[GenericFloat]
else:
    HiddenState = Any
