# Copyright 2023-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import pytest

from distilabel.utils.lists import flatten_responses


@pytest.mark.parametrize(
    "input, expected",
    [
        ([["A"], ["B"]], ["A", "B"]),
        ([["A", "B"], ["C", "D"]], ["B", "D"]),
    ],
)
def test_flatten_responses(input: List[List[str]], expected: List[str]) -> None:
    assert flatten_responses(input) == expected
