# Copyright 2023-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from distilabel.errors import DistilabelNotImplementedError
from tests.unit.conftest import DummyLLM


class TestLLM:
    def test_offline_batch_generate_raise_distilabel_not_implemented_error(
        self,
    ) -> None:
        llm = DummyLLM()

        with pytest.raises(DistilabelNotImplementedError):
            llm.offline_batch_generate()
